#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

import transaction
import zope.sqlalchemy
from pyramid.config import Configurator
from pyramid.renderers import JSONP
from pyramid_beaker import BeakerSessionFactoryConfig, session_factory_from_settings

from workspaces.notification.services.notification_info import NotificationInfo
from workspaces.notification.services.notification_service import NotificationService
from workspaces.system.schema import get_engine, get_session_factory

# ---------------------------------------------------------
#
#    S E R V I C E S
#
# ---------------------------------------------------------


def add_services(config: Configurator, session_factory: BeakerSessionFactoryConfig) -> Configurator:
    """

    :param config: Pyramid Configurator object
    :param session_factory: Pyramid Breaker session factory
    :return:
    """
    notification_info = NotificationInfo(get_tm_session(session_factory, transaction.manager))
    notification_service = NotificationService(notification_info)

    def attach_session(request):
        notification_info.session = get_tm_session(session_factory, request.tm)
        return notification_info

    # make notification_info available for use in Pyramid
    config.add_request_method(attach_session, "notification_info", reify=True)

    config.add_request_method(lambda request: notification_service, "notification_service", reify=True)
    return config


def lookup_capability_request(request):
    return request.notification_info.lookup_capability_request(request.matchdict["request_id"])


# ---------------------------------------------------------
#
#    F R A M E W O R K    S E T U P
#
# ---------------------------------------------------------


def get_tm_session(session_factory, transaction_manager):
    """
    Enable Zope's transaction manager on our session
    :param session_factory:
    :param transaction_manager:
    :return:
    """
    dbsession = session_factory()
    zope.sqlalchemy.register(dbsession, transaction_manager=transaction_manager)
    return dbsession


# ---------------------------------------------------------
#
#    M A I N    E N T R Y P O I N T
#
# ---------------------------------------------------------


def main(global_config, **settings):
    with Configurator(settings=settings) as config:
        session_factory = session_factory_from_settings(settings)
        config.set_session_factory(session_factory)
        config.add_renderer("jsonp", JSONP(param_name="callback"))

        settings["tm.manager_hook"] = "pyramid_tm.explicit_manager"

        # use pyramid_tm to hook the transaction lifecycle to the request
        config.include("pyramid_tm")

        # use pyramid_retry to retry a request when transient exceptions occur
        config.include("pyramid_retry")

        session_factory = get_session_factory(get_engine())
        config.registry["dbsession_factory"] = session_factory

        config = add_services(config, session_factory)

        config.include("pyramid_beaker")
        config.include(".routes")
        config.scan(".")
        return config.make_wsgi_app()
