#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
# pylint: disable=E0401

""" This is our workflow server API. """
import http
import json
import logging
import time
from json import JSONDecodeError
from pathlib import Path

import prometheus_client
import transaction
import zope.sqlalchemy
from pyramid.config import Configurator
from pyramid.renderers import JSONP
from pyramid.request import Request
from pyramid.response import FileResponse, Response
from pyramid.view import view_config, view_defaults
from pyramid.wsgi import wsgiapp
from pyramid_beaker import session_factory_from_settings

from workspaces.system.schema import get_engine, get_session_factory
from workspaces.workflow.schema import Workflow, WorkflowRequest, WorkflowRequestFile
from workspaces.workflow.services.recovery import MonitorRecover
from workspaces.workflow.services.workflow_info import WorkflowInfo
from workspaces.workflow.services.workflow_service import (
    WorkflowMessageHandler,
    WorkflowService,
)

logger = logging.getLogger(__name__)


prometheus_exporter = prometheus_client.make_wsgi_app()


@wsgiapp
def metrics(environ, start_response):
    return prometheus_exporter(environ, start_response)


# ---------------------------------------------------------
#
#    S E R V I C E S
#
# ---------------------------------------------------------


def lookup_workflow(request: Request) -> Workflow:
    """
    Look up the workflow definition for this workflow name.

    :param request: an HTTP request
    :return: a Workflow object
    """
    return request.info.lookup_workflow_definition(request.matchdict["name"])


def lookup_request(request: Request) -> WorkflowRequest:
    """
    Look up the workflow info for this request

    :param request: an HTTP request
    :return: a WorkflowRequest object
    """
    return request.info.lookup_workflow_request(request.matchdict["request_id"])


def lookup_file(request: Request) -> WorkflowRequestFile:
    """
    Get the file info for this request

    :param request: an HTTP request
    :return: a WorkflowRequestFile object containing filename and file content for this request, if any
    """

    for file in lookup_request(request).files:
        if file.filename == request.matchdict["filename"]:
            return file


@view_defaults(route_name="workflows", renderer="json")
class WorkflowRestService:
    """
    Top-level service for workflow requests.

    """

    def __init__(self, request: Request):
        self.request = request

    @view_config(request_method="GET")
    def list_workflows(self):
        """
        List the workflow requests that we know about.

        Audience: front-end

        :return: None
        """
        return self.request.info.all_workflows()

    @view_config(request_method="GET", route_name="workflow")
    def get_workflow(self):
        """
        Look up a workflow request

        Audience: front-end

        :return: None
        """
        return self.request.context


@view_defaults(route_name="workflow_workdir")
class WorkflowWorkingDirRestService:
    """
    Service for getting working directory files and file content
    """

    def __init__(self, request: Request):
        self.request = request

    @view_config(request_method="GET")
    def show_workdir_content(self):
        """
        List directories and files in a requested workflow's working directory

        Audience: front-end

        :return: None
        """
        requested_workflow = self.request.info.lookup_workflow_request(self.request.matchdict["request_id"])
        results_path = requested_workflow.results_dir
        parent_paths = [requested_workflow.results_dir]

        # self.request.matchdict["paths"] is a list of subpaths after dir/ in the route
        # /workflows/{name}/requests/{request_id}/dir/*path
        # Example:
        #       dir/a/b/c == ['a', 'b', 'c']
        for path in self.request.matchdict["paths"]:
            results_path += "/" + path
            parent_paths.append(results_path)

        workdir_dict = self.generate_working_directory_dict(results_path, parent_paths)

        return self.generate_response(workdir_dict)

    @view_config(request_method="GET", route_name="serve_weblog", renderer="json")
    def serve_weblog(self):
        requested_workflow = self.request.info.lookup_workflow_request(self.request.matchdict["request_id"])
        results_path = Path(requested_workflow.results_dir)

        index_path_list = list(results_path.glob("products/pipeline-*/html/index.html"))
        failed_weblog_path_list = list(results_path.glob("working/pipeline-*/html/index.html"))
        if index_path_list and len(index_path_list) == 1:
            index_path = index_path_list[0]
        elif failed_weblog_path_list and len(failed_weblog_path_list) == 1:
            # check for weblog in working directory (needed for errored or failed requests)
            index_path = failed_weblog_path_list[0]
        else:
            # otherwise, return weblog not found
            return Response(
                status_int=http.HTTPStatus.NOT_FOUND,
                json_body={
                    "resp": f"Cannot uniquely locate Weblog index file in workflow {self.request.matchdict['request_id']}."
                },
            )

        redirect_url = self.generate_url_from_path(str(index_path), str(results_path))
        return Response(status_int=http.HTTPStatus.FOUND, location=redirect_url)

    @view_config(request_method="GET", route_name="get_qa_notes", renderer="json")
    def get_qa_notes(self):
        requested_workflow = self.request.info.lookup_workflow_request(self.request.matchdict["request_id"])
        results_path = Path(requested_workflow.results_dir)

        qa_notes_list = list(results_path.glob("products/pipeline-*/html/qa_notes.html"))
        if qa_notes_list and len(qa_notes_list) == 1:
            qa_notes_path = qa_notes_list[0]
        else:
            return Response(
                status_int=http.HTTPStatus.NOT_FOUND,
                json_body={
                    "resp": f"Cannot uniquely locate QA notes file in workflow {self.request.matchdict['request_id']}."
                },
            )

        qa_notes_text = ""
        with open(qa_notes_path, "r") as qa_notes:
            qa_notes_text = qa_notes.read()

        return Response(status_int=http.HTTPStatus.OK, json_body={"resp": f"{qa_notes_text}"})

    @view_config(request_method="POST", route_name="get_qa_notes", renderer="json")
    def save_qa_notes(self):
        requested_workflow = self.request.info.lookup_workflow_request(self.request.matchdict["request_id"])
        results_path = Path(requested_workflow.results_dir)

        qa_notes_list = list(results_path.glob("products/pipeline-*/html/qa_notes.html"))
        if qa_notes_list and len(qa_notes_list) == 1:
            qa_notes_path = qa_notes_list[0]
        else:
            return Response(
                status_int=http.HTTPStatus.NOT_FOUND,
                json_body={
                    "resp": f"Cannot uniquely locate QA notes file in workflow {self.request.matchdict['request_id']}."
                },
            )

        # If an exception is thrown with the qa_notes file open, its contents will be wiped
        if "edits" not in self.request.json_body:
            return Response(
                status_int=http.HTTPStatus.OK,
                json_body={
                    "resp": f"No edits were made to QA notes file in workflow {self.request.matchdict['request_id']}."
                },
            )

        with open(qa_notes_path, "w") as qa_notes:
            # sanitize input before writing/persisting
            # \\u0000 is an invalid character that is incompatible with postgres json columns
            # from StackOverflow: https://stackoverflow.com/questions/63092267/how-to-handle-api-responsesjson-containing-x00-or-u0000-in-its-data-and-s
            edits = json.loads(json.dumps(self.request.json_body["edits"])).replace("\\u0000", "")
            qa_notes.write(edits)

        return Response(
            status_int=http.HTTPStatus.OK,
            json_body={"resp": f"Edits made to QA notes file in workflow {self.request.matchdict['request_id']}."},
        )

    @view_config(request_method="GET", route_name="serve_carta_wrapper", renderer="json")
    def serve_carta_wrapper(self):
        """
        Dish up some HTML containing the CARTA URL in a frame.

        :return:
        """

        path = Path(f"/lustre/aoc/cluster/pipeline/docker/workspaces/html/{self.request.matchdict['request_id']}")
        carta_html_file = list(path.iterdir())[0]
        return FileResponse(carta_html_file, request=self.request, content_type="text/html")

    @view_config(route_name="get_healthcheck", renderer="json")
    def get_healthcheck(self) -> Response:
        """
        Route to let capabilities check that workflow is ready

        :return:
        """
        return Response(
            status=http.HTTPStatus.OK,
            json_body={"healthcheck": f"Workflow service returned {http.HTTPStatus.OK}"},
        )

    def generate_working_directory_dict(self, results_path, parent_paths) -> dict:
        workdir_dict = {}
        # add 'root' key with requested_workflow.results_dir as the value
        workdir_dict.update({"root": parent_paths[0]})

        # get the last item from the parent_paths list
        # the last item is always the current path
        current_path = parent_paths.pop()
        workdir_dict.update({"current_path": current_path})

        # add 'parents' key with the remaining items in the parent_paths list
        workdir_dict.update({"parents": parent_paths})

        # if the path is a file, return workdir_dict without directory content info
        if Path(results_path).is_file():
            return workdir_dict

        # add 'content' key for housing directory and file information
        workdir_dict.update({"content": {}})
        for key in Path(results_path).iterdir():
            content_key = workdir_dict["content"]

            # check if url needs a slash to divide paths
            divider = ("/", "")[self.request.current_route_url().endswith("/")]
            content_key.update({key.name: {"url": self.request.current_route_url() + divider + key.name}})

            # add full path for content
            content_key[key.name].update({"full_path": key.absolute().__str__()})

            # determine if content is a directory or a file
            content_type = ("directory", "file")[Path.is_file(key)]
            content_key[key.name].update({"content_type": content_type})
        return workdir_dict

    def generate_response(self, workdir_dict):
        # check if current path is a directory
        if Path(workdir_dict.get("current_path")).is_dir():
            # if it is a directory, create a json object
            workdir_json = json.dumps(workdir_dict, indent=2)
            # create response with the json object as the body
            response = Response(body=workdir_json, request=self.request, content_type="text/json")
        else:
            # if it is not a directory, serve the static file
            response = FileResponse(
                workdir_dict.get("current_path"),
                request=self.request,
            )
        return response

    def generate_url_from_path(self, root_path, results_path):
        current_url = self.request.current_route_url()
        return current_url.replace("/weblog", "/dir") + root_path.replace(results_path, "")

    def generate_qa_notes_path(self, root_path, results_path):
        current_url = self.request.current_route_url()
        return current_url.replace("/qa_notes", "/dir") + root_path.replace(results_path, "")


@view_defaults(route_name="workflow_request", renderer="json")
class WorkflowRequestRestService:
    """
    Services for the user-submitted files attached to workflows.
    """

    def __init__(self, request: Request):
        self.request = request

    @view_config(request_method="GET")
    def get_workflow_request(self):
        return self.request.context

    @view_config(request_method="POST", route_name="create_workflow_request")
    def create_workflow_request(self):
        """
        Create a new workflow request from the name/arguments supplied.

        Audience: front-end and CLI
        :return:
        """

        # Decode the body for details about the incoming request
        try:
            argument_body = self.request.json_body
        except JSONDecodeError:
            # Most common case: Empty body for simple requests, continue with an empty dict
            argument_body = {}

        return self.request.info.create_workflow_request(self.request.context.workflow_name, argument_body)

    @view_config(request_method="POST", route_name="submit_workflow_request")
    def submit_workflow(self):
        """
        Submit this workflow request for processing.

        Audience: front-end and CLI
        :return:
        """
        print(f"Submitting workflow {self.request.context}")

        return self.request.workflows.execute(self.request.context)

    @view_config(request_method="POST", route_name="create_and_submit_workflow_request")
    def create_and_submit(self):
        """
        Create a new workflow request from the name/arguments supplied, then submit the request.

        Audience: front-end and CLI
        :return:
        """
        self.request.context = self.create_workflow_request()
        return self.submit_workflow()

    @view_config(request_method="POST", route_name="ingest_workflow_result")
    def ingest(self):
        """
        Ingest specified workflow request's results into NGAS and archive

        :return:
        """
        print(f"Ingesting results for workflow request {self.request.context}")

        # 1. retrieve metadata.json for workflow request
        self.request.matchdict["filename"] = "metadata.json"
        file = lookup_file(request=self.request)

        # 2. create ingestion workflow request
        ingest_type = "ingest_cal" if "calibration" in self.request.matchdict["name"] else "ingest_image"
        ingest_request = self.request.info.create_workflow_request(
            workflow=ingest_type,
            argument={"parent_wf_request_id": self.request.matchdict["request_id"]},
        )
        # 3. attach metadata.json to ingestion wf request
        self.request.workflows.attach_file_to_request(
            request=ingest_request, filename=file.filename, content=file.content
        )
        # 4. submit ingestion workflow request
        self.request.workflows.execute(ingest_request)

    @view_config(request_method="POST", route_name="ingest_vlass_products")
    def vlass_ingest(self):
        """
        VLASS ingestion requires different setup for WS ingestion. Ingest VLASS products into NGAS and the archive.

        :return:
        """
        vlass_type = self.request.matchdict["vlass_type"]

        ingestion_request = self.request.info.create_workflow_request(
            workflow="ingest_" + vlass_type,
            argument=self.request.json_body,
        )
        self.request.workflows.execute(ingestion_request)
        return Response(status_code=http.HTTPStatus.OK, json_body=ingestion_request.__json__())

    @view_config(request_method="POST", route_name="abort_workflow_request")
    def abort(self):
        """
        Given the ID of a workflow request, abort the execution of the request if it is running
        """
        request_id = self.request.matchdict["request_id"]
        self.request.workflows.abort_running_workflow(request_id)

        return Response(
            status_code=http.HTTPStatus.OK,
            body=f"SUCCESS: Sent signal to abort running workflow request #{request_id}.",
        )

    @view_config(request_method="POST", route_name="qa_fail")
    def qa_fail(self):
        """
        Perform qa_fail on specified workflow

        :return:
        """
        print(f"QA Failing workflow request #{self.request.context}")

        # 1. retrieve metadata.json for workflow request
        self.request.matchdict["filename"] = "metadata.json"
        file = lookup_file(request=self.request)

        # 2. create qa fail workflow request
        fail_request = self.request.info.create_workflow_request(
            workflow="qa_fail",
            argument={"parent_wf_request_id": self.request.matchdict["request_id"]},
        )
        # 3. attach metadata.json to fail wf request
        self.request.workflows.attach_file_to_request(
            request=fail_request, filename=file.filename, content=file.content
        )
        # 4. submit ingestion workflow request
        self.request.workflows.execute(fail_request)

    @view_config(request_method="POST", route_name="send_carta_url_to_aat")
    def send_carta_url_to_aat(self):
        """
        Pyramid view that sends a given CARTA URL to the AAT's AMQP exchange via the WorkflowService

        return: 200 Response all day, every day
        """
        request_id = self.request.matchdict["request_id"]
        carta_url = self.request.json_body["carta_url"]
        self.request.workflows.send_carta_url_to_aat(request_id, carta_url)

        return Response(
            status_code=http.HTTPStatus.OK,
            body=f"SUCCESS: Sent CARTA URL {carta_url} to AAT system",
        )

    @view_config(request_method="POST", route_name="send_carta_url")
    def send_carta_url(self):
        """
        Pyramid view that sends a given CARTA URL to the Workspaces system

        return: 200 (OK) HTTP Response
        """
        request_id = self.request.matchdict["request_id"]
        carta_url = self.request.json_body["carta_url"]
        self.request.workflows.send_carta_url(request_id, carta_url)

        return Response(
            status_code=http.HTTPStatus.OK,
            body=f"SUCCESS: Sent CARTA URL {carta_url} to Workspaces System",
        )

    @view_config(request_method="POST", route_name="message_archive")
    def message_archive(self):
        """
        Pyramid view that sends a message to the Workspaces system

        return: 200 (OK) HTTP Response
        """
        body = self.request.json_body

        identifier = int(body["request_id"]) if "request_id" in body else body["project_code"]
        msg_type = self.request.matchdict["msg_type"]

        additional = None

        if "project_code" in body and isinstance(identifier, int):
            additional = body["project_code"]
        elif "telescope" in body and isinstance(identifier, str):
            additional = body["telescope"]

        self.request.workflows.message_archive(identifier, msg_type, additional)

        return Response(
            status_code=http.HTTPStatus.OK,
            body=f"SUCCESS for identifier '{identifier}': Sent '{msg_type}' to AAT for this workflow's SDM",
        )

    @view_config(request_method="POST", route_name="announce_qa")
    def announce_qa(self):
        """
        Announce QA ready, pass, or fail

        :return:
        """
        request_id = self.request.matchdict["request_id"]
        msg_type = self.request.matchdict["msg_type"]
        version = self.request.json_body["capability_version"]
        self.request.workflows.announce_qa(request_id, msg_type, version)

        return Response(
            status_code=http.HTTPStatus.OK,
            body=f"SUCCESS: Sent {msg_type} to Workspaces System",
        )

    @view_config(request_method="POST", route_name="force_fail")
    def send_force_fail(self):
        """
        Pyramid view that sends a fail message for the specified workflow

        :return:
        """
        request_id = self.request.matchdict["request_id"]
        self.request.workflows.send_forced_fail(request_id)

    @view_config(request_method="GET", route_name="workflow_request_htcondor_id", renderer="json")
    def get_request_htcondor_id(self):
        """
        Pyramid view that gives back the HTCondor job ID for a given workflow request

        :return: HTTP response with HTCondor job ID in the body
        """
        requested_workflow = self.request.info.lookup_workflow_request(self.request.matchdict["request_id"])

        return Response(json_body={"htcondor_job_id": str(requested_workflow.htcondor_job_id)})

    @view_config(request_method="GET", route_name="list_stale_requests")
    def get_stale_requests(self):
        """
        Pyramid view that returns a list of workflow request processing directories that
        have completed processing and are marked for annihilation

        :return:  HTTP Response with list of directories to annihilate
        """
        dir_list = []
        keep = self.request.matchdict["days"]

        req_list = self.request.info.lookup_stale_requests(int(keep))

        for r in req_list:
            if r.results_dir:
                dir_list.append(r.results_dir)

        # return list as comma seperated string
        return Response(json_body={"stale_dirs": str(",".join(dir_list))})

    @view_config(request_method="PUT", route_name="set_request_cleaned")
    def set_request_cleaned(self):
        """
        Pyramid view to set the workflow request cleaned flag upon directory annihilation

        :return: HTTP Response
        """

        directory = self.request.json_body["directory"]
        update_to = eval(self.request.json_body["update_to"])
        req = self.request.info.lookup_workflow_request_by_directory(directory)
        if req:
            self.request.info.update_request_cleaned(req, update_to)
            return Response(
                status_code=http.HTTPStatus.OK,
                body=f"SUCCESS: Updated workflow request #{req.workflow_request_id} to {update_to}",
            )
        else:
            return Response(
                status_code=http.HTTPStatus.OK,
                body=f"SKIPPED: Directory {directory} does not belong to the current system.",
            )


@view_defaults(route_name="workflow_request_files", renderer="json")
class WorkflowFilesRestService:
    """
    Services for the user-submitted files attached to workflows.
    """

    def __init__(self, request):
        self.request = request

    @view_config(request_method="PUT", route_name="workflow_request_file")
    def add_file(self):
        """
        Add a file to this workflow request.

        Audience: front-end and CLI
        """
        logger.info(
            "Adding file %s to %s",
            self.request.matchdict["filename"],
            self.request.context,
        )
        file = self.request.workflows.attach_file_to_request(
            request=self.request.context,
            filename=self.request.matchdict["filename"],
            content=self.request.body,
        )
        return file

    @view_config(request_method="GET")
    def get_files(self):
        """
        Get the files associated with this workflow request.

        Audience: front-end
        :return:
        """
        return self.request.context["files"]

    @view_config(
        request_method="GET",
        accept="application/octet-stream",
        route_name="workflow_request_file",
    )
    def get_file_text(self):
        """
        Get the text contents of this file

        Audience: ???
        :return:
        """
        self.request.response.content_type = "application/octet-stream"
        self.request.response.body = lookup_file(self.request).content
        return self.request.response

    @view_config(
        request_method="GET",
        accept="application/json",
        route_name="workflow_request_file",
    )
    def get_file_json(self):
        """
        Get a JSON representation of this file

        Audience: ???
        :return:
        """
        return self.request.context


# ---------------------------------------------------------
#
#    F R A M E W O R K    S E T U P
#
# ---------------------------------------------------------


def get_tm_session(session_factory, transaction_manager):
    """
    Enable Zope's transaction manager on our session
    :param session_factory:
    :param transaction_manager:
    :return:
    """
    dbsession = session_factory()
    zope.sqlalchemy.register(dbsession, transaction_manager=transaction_manager)
    return dbsession


# ---------------------------------------------------------
#
#    P R O M E T H E U S   S E T U P
#
# ---------------------------------------------------------

http_requests = None


def prometheus_route_timing_factory(handler, registry):
    # if timing support is enabled, return a wrapper
    global http_requests
    # Prometheus logging
    http_requests = prometheus_client.Summary("http_request_timing", "HTTP Requests", ["status_code", "route_name"])

    def prometheus_route_timer(request):
        start = time.time()
        response = handler(request)
        end = time.time()
        if request.matched_route:
            http_requests.labels(status_code=response.status_code, route_name=request.matched_route.name).observe(
                end - start
            )
        return response

    return prometheus_route_timer


# ---------------------------------------------------------
#
#    M A I N    E N T R Y P O I N T
#
# ---------------------------------------------------------


def main(global_config, **settings):
    with Configurator(settings=settings) as config:
        session_factory = session_factory_from_settings(settings)
        config.set_session_factory(session_factory)
        config.add_renderer("jsonp", JSONP(param_name="callback"))

        settings["tm.manager_hook"] = "pyramid_tm.explicit_manager"

        # use pyramid_tm to hook the transaction lifecycle to the request
        config.include("pyramid_tm")

        # use pyramid_retry to retry a request when transient exceptions occur
        config.include("pyramid_retry")

        session_factory = get_session_factory(get_engine())
        config.registry["dbsession_factory"] = session_factory

        # we need to build a workflow_info here for the message handler, but
        # we won't use it anywhere else, we will make new ones per-request
        workflow_info = WorkflowInfo(get_tm_session(session_factory, transaction.manager))
        message_handler = WorkflowMessageHandler(workflow_info)
        workflow_recover = MonitorRecover(workflow_info)

        def create_workflow_info(request):
            session = get_tm_session(session_factory, request.tm)
            return WorkflowInfo(session)

        config.add_request_method(create_workflow_info, "info", reify=True)

        # make workflow_service available for use in Pyramid
        config.add_request_method(
            lambda r: WorkflowService(r.info, message_handler, workflow_recover),
            "workflows",
            reify=True,
        )

        # GET  /workflows                            <- list of workflows
        # GET  /workflows/null                       <- info about the null workflow
        # POST /workflows/null/requests/create       <- create a request for the null workflow
        # PUT  /workflows/requests/23/files/foo.txt  <- attach foo.txt to request #23 on workflow null
        # POST /workflows/requests/23/submit         <- launch request #23

        config.add_route("workflows", "/workflows")
        config.add_route("workflow", "/workflows/{name}", factory=lookup_workflow)

        # parameters are passed with json: {workflow:<name>, args: <args>}
        config.add_route(
            "create_workflow_request",
            "/workflows/{name}/requests/create",
            factory=lookup_workflow,
        )
        config.add_route(
            "create_and_submit_workflow_request",
            "/workflows/{name}/requests/create-and-submit",
            factory=lookup_workflow,
        )
        config.add_route(
            "workflow_request",
            "/workflows/{name}/requests/{request_id}",
            factory=lookup_request,
        )
        config.add_route(
            "workflow_workdir",
            "/workflows/{name}/requests/{request_id}/dir/*paths",
            factory=lookup_request,
        )
        config.add_route(
            "serve_weblog",
            "/workflows/{name}/requests/{request_id}/weblog",
            factory=lookup_request,
        )
        config.add_route(
            "get_qa_notes",
            "/workflows/{name}/requests/{request_id}/qa_notes",
            factory=lookup_request,
        )
        config.add_route(
            "serve_carta_wrapper",
            "/workflows/{name}/requests/{request_id}/html",
            factory=lookup_request,
        )
        config.add_route(
            "workflow_request_files",
            "/workflows/{name}/requests/{request_id}/files",
            factory=lookup_request,
        )
        config.add_route(
            "workflow_request_htcondor_id",
            "/workflows/{name}/requests/{request_id}/htcondor_id",
            factory=lookup_request,
        )
        config.add_route(
            "workflow_request_file",
            "/workflows/{name}/requests/{request_id}/files/{filename}",
            factory=lookup_request,
        )

        # parameters are passed with json: {files:{<name>}}.
        # Use empty braces for no additional files
        config.add_route(
            "submit_workflow_request",
            "/workflows/{name}/requests/{request_id}/submit",
            factory=lookup_request,
        )
        # Set off ingestion follow-on workflow
        config.add_route(
            "ingest_workflow_result",
            "/workflows/{name}/requests/{request_id}/ingest",
            factory=lookup_request,
        )
        # parameters are passed with json: ex. {cal_spl: <cal spl>, cache_path: <cache directory>}
        config.add_route(
            "ingest_vlass_products",
            "workflows/requests/vlass_ingest/{vlass_type}",
        )
        # Use this route to abort a running workflow
        config.add_route(
            "abort_workflow_request",
            "/workflows/requests/{request_id}/abort",
            factory=lookup_request,
        )

        # Use this route to send a message to the AAT system
        config.add_route(
            "message_archive",
            "/workflows/requests/message-archive/{msg_type}",
        )

        # Use this route to send a CARTA envoy URL to the AAT system
        config.add_route(
            "send_carta_url_to_aat",
            "/workflows/carta/requests/{request_id}/send-url-to-aat",
            factory=lookup_request,
        )
        # Use this route to send a CARTA envoy URL to the Workspaces system
        config.add_route(
            "send_carta_url",
            "/workflows/carta/requests/{request_id}/url",
            factory=lookup_request,
        )

        # Use this to force fail a workflow request
        config.add_route(
            "force_fail",
            "/workflows/{name}/requests/{request_id}/fail",
            factory=lookup_request,
        )

        config.add_route(
            "qa_fail",
            "/workflows/{name}/requests/{request_id}/qa_fail",
            factory=lookup_request,
        )

        # Use this route to send a QA ready, pass, or fail announcement
        config.add_route(
            "announce_qa",
            "/workflows/requests/{request_id}/qa/{msg_type}",
            factory=lookup_request,
        )

        # route for retrieving a list of requests viable for clean up
        config.add_route(
            "list_stale_requests",
            "/workflows/requests/stale/{days}",
        )

        # route for setting cleaned status via annihilator
        config.add_route(
            "set_request_cleaned",
            "/workflows/requests/cleaned",
        )

        # Route for healthcheck when bringing up containers
        config.add_route(
            "get_healthcheck",
            "/healthcheck",
        )

        config.add_view(metrics, name="metrics")
        config.add_route("metrics", "metrics")

        config.include("pyramid_beaker")
        config.scan(".")

        # add the prometheus timing metric
        config.add_tween("workflow.server.prometheus_route_timing_factory")

        return config.make_wsgi_app()
